@@ -4,16 +4,18 @@ from collections import deque, namedtuple
 from contextlib import contextmanager
 from decimal import Decimal
 from io import BytesIO
+from typing import Any
 from unittest import TestCase, mock
 from urllib.request import pathname2url
 import json
 import os
 import sys
 import tempfile
-import unittest
 import warnings
 
-import attr
+from attrs import define, field
+from referencing.jsonschema import DRAFT202012
+import referencing.exceptions
 
 from jsonschema import (
     FormatChecker,
@@ -22,7 +24,6 @@ from jsonschema import (
     protocols,
     validators,
 )
-from jsonschema.tests._helpers import bug
 
 
 def fail(validator, errors, instance, schema):
@@ -38,6 +39,11 @@ class TestCreateAndExtend(TestCase):
             validators._META_SCHEMAS,
             dict(validators._META_SCHEMAS),
         )
+        self.addCleanup(
+            self.assertEqual,
+            validators._VALIDATORS,
+            dict(validators._VALIDATORS),
+        )
 
         self.meta_schema = {"$id": "some://meta/schema"}
         self.validators = {"fail": fail}
@@ -108,6 +114,7 @@ class TestCreateAndExtend(TestCase):
             version="my version",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, "something")
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
         self.assertEqual(Validator.__name__, "MyVersionValidator")
         self.assertEqual(Validator.__qualname__, "MyVersionValidator")
 
@@ -117,6 +124,7 @@ class TestCreateAndExtend(TestCase):
             version="my version",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, "something")
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
         self.assertEqual(
             repr(Validator({})),
             "MyVersionValidator(schema={}, format_checker=None)",
@@ -128,6 +136,7 @@ class TestCreateAndExtend(TestCase):
             version="my version",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, "something")
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
         self.assertEqual(
             repr(Validator({"a": list(range(1000))})), (
                 "MyVersionValidator(schema={'a': [0, 1, 2, 3, 4, 5, ...]}, "
@@ -148,6 +157,7 @@ class TestCreateAndExtend(TestCase):
             version="foo-bar",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, "something")
+        self.addCleanup(validators._VALIDATORS.pop, "foo-bar")
         self.assertEqual(Validator.__qualname__, "FooBarValidator")
 
     def test_if_a_version_is_not_provided_it_is_not_registered(self):
@@ -165,6 +175,7 @@ class TestCreateAndExtend(TestCase):
             id_of=lambda s: s.get("id", ""),
         )
         self.addCleanup(validators._META_SCHEMAS.pop, meta_schema_key)
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
 
         self.assertIn(meta_schema_key, validators._META_SCHEMAS)
 
@@ -177,6 +188,7 @@ class TestCreateAndExtend(TestCase):
             version="my version",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, meta_schema_key)
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
 
         self.assertIn(meta_schema_key, validators._META_SCHEMAS)
 
@@ -273,6 +285,23 @@ class TestCreateAndExtend(TestCase):
         Derived = validators.extend(Original)
         self.assertEqual(Derived.ID_OF(Derived.META_SCHEMA), correct_id)
 
+    def test_extend_applicable_validators(self):
+        """
+        Extending a validator preserves its notion of applicable validators.
+        """
+
+        schema = {
+            "$defs": {"test": {"type": "number"}},
+            "$ref": "#/$defs/test",
+            "maximum": 1
+        }
+
+        draft4 = validators.Draft4Validator(schema)
+        self.assertTrue(draft4.is_valid(37))  # as $ref ignores siblings
+
+        Derived = validators.extend(validators.Draft4Validator)
+        self.assertTrue(Derived(schema).is_valid(37))
+
 
 class TestValidationErrorMessages(TestCase):
     def message_for(self, instance, schema, *args, **kwargs):
@@ -1178,7 +1207,7 @@ class TestValidationErrorDetails(TestCase):
         ref, schema = "someRef", {"additionalProperties": {"type": "integer"}}
         validator = validators.Draft7Validator(
             {"$ref": ref},
-            resolver=validators.RefResolver("", {}, store={ref: schema}),
+            resolver=validators._RefResolver("", {}, store={ref: schema}),
         )
         error, = validator.iter_errors({"foo": "notAnInteger"})
 
@@ -1513,7 +1542,7 @@ class MetaSchemaTestsMixin:
         )
 
 
-class ValidatorTestMixin(MetaSchemaTestsMixin, object):
+class ValidatorTestMixin(MetaSchemaTestsMixin):
     def test_it_implements_the_validator_protocol(self):
         self.assertIsInstance(self.Validator({}), protocols.Validator)
 
@@ -1528,31 +1557,25 @@ class ValidatorTestMixin(MetaSchemaTestsMixin, object):
     def test_non_existent_properties_are_ignored(self):
         self.Validator({object(): object()}).validate(instance=object())
 
-    def test_it_creates_a_ref_resolver_if_not_provided(self):
-        self.assertIsInstance(
-            self.Validator({}).resolver,
-            validators.RefResolver,
+    def test_evolve(self):
+        schema, format_checker = {"type": "integer"}, FormatChecker()
+        original = self.Validator(
+            schema,
+            format_checker=format_checker,
+        )
+        new = original.evolve(
+            schema={"type": "string"},
+            format_checker=self.Validator.FORMAT_CHECKER,
         )
 
-    def test_it_delegates_to_a_ref_resolver(self):
-        ref, schema = "someCoolRef", {"type": "integer"}
-        resolver = validators.RefResolver("", {}, store={ref: schema})
-        validator = self.Validator({"$ref": ref}, resolver=resolver)
-
-        with self.assertRaises(exceptions.ValidationError):
-            validator.validate(None)
-
-    def test_evolve(self):
-        ref, schema = "someCoolRef", {"type": "integer"}
-        resolver = validators.RefResolver("", {}, store={ref: schema})
-
-        validator = self.Validator(schema, resolver=resolver)
-        new = validator.evolve(schema={"type": "string"})
-
-        expected = self.Validator({"type": "string"}, resolver=resolver)
+        expected = self.Validator(
+            {"type": "string"},
+            format_checker=self.Validator.FORMAT_CHECKER,
+            _resolver=new._resolver,
+        )
 
         self.assertEqual(new, expected)
-        self.assertNotEqual(new, validator)
+        self.assertNotEqual(new, original)
 
     def test_evolve_with_subclass(self):
         """
@@ -1565,10 +1588,10 @@ class ValidatorTestMixin(MetaSchemaTestsMixin, object):
         """
 
         with self.assertWarns(DeprecationWarning):
-            @attr.s
+            @define
             class OhNo(self.Validator):
-                foo = attr.ib(factory=lambda: [1, 2, 3])
-                _bar = attr.ib(default=37)
+                foo = field(factory=lambda: [1, 2, 3])
+                _bar = field(default=37)
 
         validator = OhNo({}, bar=12)
         self.assertEqual(validator.foo, [1, 2, 3])
@@ -1577,24 +1600,6 @@ class ValidatorTestMixin(MetaSchemaTestsMixin, object):
         self.assertEqual(new.foo, [1, 2, 3])
         self.assertEqual(new._bar, 12)
 
-    def test_it_delegates_to_a_legacy_ref_resolver(self):
-        """
-        Legacy RefResolvers support only the context manager form of
-        resolution.
-        """
-
-        class LegacyRefResolver:
-            @contextmanager
-            def resolving(this, ref):
-                self.assertEqual(ref, "the ref")
-                yield {"type": "integer"}
-
-        resolver = LegacyRefResolver()
-        schema = {"$ref": "the ref"}
-
-        with self.assertRaises(exceptions.ValidationError):
-            self.Validator(schema, resolver=resolver).validate(None)
-
     def test_is_type_is_true_for_valid_type(self):
         self.assertTrue(self.Validator({}).is_type("foo", "string"))
 
@@ -1647,7 +1652,7 @@ class ValidatorTestMixin(MetaSchemaTestsMixin, object):
             elif value == "bad":
                 raise bad
             else:  # pragma: no cover
-                self.fail("What is {}? [Baby Don't Hurt Me]".format(value))
+                self.fail(f"What is {value}? [Baby Don't Hurt Me]")
 
         validator = self.Validator(
             {"format": "foo"}, format_checker=checker,
@@ -1755,6 +1760,37 @@ class ValidatorTestMixin(MetaSchemaTestsMixin, object):
             with self.assertRaises(exceptions.ValidationError):
                 validator.validate(instance)
 
+    def test_it_creates_a_ref_resolver_if_not_provided(self):
+        with self.assertWarns(DeprecationWarning):
+            resolver = self.Validator({}).resolver
+        self.assertIsInstance(resolver, validators._RefResolver)
+
+    def test_it_upconverts_from_deprecated_RefResolvers(self):
+        ref, schema = "someCoolRef", {"type": "integer"}
+        resolver = validators._RefResolver("", {}, store={ref: schema})
+        validator = self.Validator({"$ref": ref}, resolver=resolver)
+
+        with self.assertRaises(exceptions.ValidationError):
+            validator.validate(None)
+
+    def test_it_upconverts_from_yet_older_deprecated_legacy_RefResolvers(self):
+        """
+        Legacy RefResolvers support only the context manager form of
+        resolution.
+        """
+
+        class LegacyRefResolver:
+            @contextmanager
+            def resolving(this, ref):
+                self.assertEqual(ref, "the ref")
+                yield {"type": "integer"}
+
+        resolver = LegacyRefResolver()
+        schema = {"$ref": "the ref"}
+
+        with self.assertRaises(exceptions.ValidationError):
+            self.Validator(schema, resolver=resolver).validate(None)
+
 
 class AntiDraft6LeakMixin:
     """
@@ -1771,18 +1807,14 @@ class AntiDraft6LeakMixin:
             self.Validator.check_schema(False)
         self.assertIn("False is not of type", str(e.exception))
 
-    @unittest.skip(bug(523))
     def test_True_is_not_a_schema_even_if_you_forget_to_check(self):
-        resolver = validators.RefResolver("", {})
         with self.assertRaises(Exception) as e:
-            self.Validator(True, resolver=resolver).validate(12)
+            self.Validator(True).validate(12)
         self.assertNotIsInstance(e.exception, exceptions.ValidationError)
 
-    @unittest.skip(bug(523))
     def test_False_is_not_a_schema_even_if_you_forget_to_check(self):
-        resolver = validators.RefResolver("", {})
         with self.assertRaises(Exception) as e:
-            self.Validator(False, resolver=resolver).validate(12)
+            self.Validator(False).validate(12)
         self.assertNotIsInstance(e.exception, exceptions.ValidationError)
 
 
@@ -1856,7 +1888,7 @@ class TestLatestValidator(TestCase):
     def test_ref_resolvers_may_have_boolean_schemas_stored(self):
         ref = "someCoolRef"
         schema = {"$ref": ref}
-        resolver = validators.RefResolver("", {}, store={ref: False})
+        resolver = validators._RefResolver("", {}, store={ref: False})
         validator = validators._LATEST_VERSION(schema, resolver=resolver)
 
         with self.assertRaises(exceptions.ValidationError):
@@ -2103,6 +2135,63 @@ class TestValidate(TestCase):
         self.assertIn("12 is less than the minimum of 20", str(e.exception))
 
 
+class TestThreading(TestCase):
+    """
+    Threading-related functionality tests.
+
+    jsonschema doesn't promise thread safety, and its validation behavior
+    across multiple threads may change at any time, but that means it isn't
+    safe to share *validators* across threads, not that anytime one has
+    multiple threads that jsonschema won't work (it certainly is intended to).
+
+    These tests ensure that this minimal level of functionality continues to
+    work.
+    """
+
+    def test_validation_across_a_second_thread(self):
+        failed = []
+
+        def validate():
+            try:
+                validators.validate(instance=37, schema=True)
+            except:  # pragma: no cover  # noqa: E722
+                failed.append(sys.exc_info())
+
+        validate()  # just verify it succeeds
+
+        from threading import Thread
+        thread = Thread(target=validate)
+        thread.start()
+        thread.join()
+        self.assertEqual((thread.is_alive(), failed), (False, []))
+
+
+class TestReferencing(TestCase):
+    def test_registry_with_retrieve(self):
+        def retrieve(uri):
+            return DRAFT202012.create_resource({"type": "integer"})
+
+        registry = referencing.Registry(retrieve=retrieve)
+        schema = {"$ref": "https://example.com/"}
+        validator = validators.Draft202012Validator(schema, registry=registry)
+
+        self.assertEqual(
+            (validator.is_valid(12), validator.is_valid("foo")),
+            (True, False),
+        )
+
+    def test_custom_registries_do_not_autoretrieve_remote_resources(self):
+        registry = referencing.Registry()
+        schema = {"$ref": "https://example.com/"}
+        validator = validators.Draft202012Validator(schema, registry=registry)
+
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter("always")
+            with self.assertRaises(referencing.exceptions.Unresolvable):
+                validator.validate(12)
+        self.assertFalse(w)
+
+
 class TestRefResolver(TestCase):
 
     base_uri = ""
@@ -2112,7 +2201,7 @@ class TestRefResolver(TestCase):
     def setUp(self):
         self.referrer = {}
         self.store = {self.stored_uri: self.stored_schema}
-        self.resolver = validators.RefResolver(
+        self.resolver = validators._RefResolver(
             self.base_uri, self.referrer, self.store,
         )
 
@@ -2132,7 +2221,7 @@ class TestRefResolver(TestCase):
 
     def test_it_resolves_local_refs_with_id(self):
         schema = {"id": "http://bar/schema#", "a": {"foo": "bar"}}
-        resolver = validators.RefResolver.from_schema(
+        resolver = validators._RefResolver.from_schema(
             schema,
             id_of=lambda schema: schema.get("id", ""),
         )
@@ -2153,7 +2242,7 @@ class TestRefResolver(TestCase):
         ref = "http://bar#baz"
         schema = {"baz": 12}
 
-        if "requests" in sys.modules:
+        if "requests" in sys.modules:  # pragma: no cover
             self.addCleanup(
                 sys.modules.__setitem__, "requests", sys.modules["requests"],
             )
@@ -2166,7 +2255,7 @@ class TestRefResolver(TestCase):
         ref = "http://bar#baz"
         schema = {"baz": 12}
 
-        if "requests" in sys.modules:
+        if "requests" in sys.modules:  # pragma: no cover
             self.addCleanup(
                 sys.modules.__setitem__, "requests", sys.modules["requests"],
             )
@@ -2189,13 +2278,13 @@ class TestRefResolver(TestCase):
             self.addCleanup(os.remove, tempf.name)
             json.dump({"foo": "bar"}, tempf)
 
-        ref = "file://{}#foo".format(pathname2url(tempf.name))
+        ref = f"file://{pathname2url(tempf.name)}#foo"
         with self.resolver.resolving(ref) as resolved:
             self.assertEqual(resolved, "bar")
 
     def test_it_can_construct_a_base_uri_from_a_schema(self):
         schema = {"id": "foo"}
-        resolver = validators.RefResolver.from_schema(
+        resolver = validators._RefResolver.from_schema(
             schema,
             id_of=lambda schema: schema.get("id", ""),
         )
@@ -2212,7 +2301,7 @@ class TestRefResolver(TestCase):
 
     def test_it_can_construct_a_base_uri_from_a_schema_without_id(self):
         schema = {}
-        resolver = validators.RefResolver.from_schema(schema)
+        resolver = validators._RefResolver.from_schema(schema)
         self.assertEqual(resolver.base_uri, "")
         self.assertEqual(resolver.resolution_scope, "")
         with resolver.resolving("") as resolved:
@@ -2227,7 +2316,7 @@ class TestRefResolver(TestCase):
 
         schema = {"foo": "bar"}
         ref = "foo://bar"
-        resolver = validators.RefResolver("", {}, handlers={"foo": handler})
+        resolver = validators._RefResolver("", {}, handlers={"foo": handler})
         with resolver.resolving(ref) as resolved:
             self.assertEqual(resolved, schema)
 
@@ -2241,7 +2330,7 @@ class TestRefResolver(TestCase):
                 self.fail("Response must not have been cached!")
 
         ref = "foo://bar"
-        resolver = validators.RefResolver(
+        resolver = validators._RefResolver(
             "", {}, cache_remote=True, handlers={"foo": handler},
         )
         with resolver.resolving(ref):
@@ -2259,7 +2348,7 @@ class TestRefResolver(TestCase):
                 self.fail("Handler called twice!")
 
         ref = "foo://bar"
-        resolver = validators.RefResolver(
+        resolver = validators._RefResolver(
             "", {}, cache_remote=False, handlers={"foo": handler},
         )
         with resolver.resolving(ref):
@@ -2272,19 +2361,76 @@ class TestRefResolver(TestCase):
             raise error
 
         ref = "foo://bar"
-        resolver = validators.RefResolver("", {}, handlers={"foo": handler})
-        with self.assertRaises(exceptions.RefResolutionError) as err:
+        resolver = validators._RefResolver("", {}, handlers={"foo": handler})
+        with self.assertRaises(exceptions._RefResolutionError) as err:
             with resolver.resolving(ref):
                 self.fail("Shouldn't get this far!")  # pragma: no cover
-        self.assertEqual(err.exception, exceptions.RefResolutionError(error))
+        self.assertEqual(err.exception, exceptions._RefResolutionError(error))
 
     def test_helpful_error_message_on_failed_pop_scope(self):
-        resolver = validators.RefResolver("", {})
+        resolver = validators._RefResolver("", {})
         resolver.pop_scope()
-        with self.assertRaises(exceptions.RefResolutionError) as exc:
+        with self.assertRaises(exceptions._RefResolutionError) as exc:
             resolver.pop_scope()
         self.assertIn("Failed to pop the scope", str(exc.exception))
 
+    def test_pointer_within_schema_with_different_id(self):
+        """
+        See #1085.
+        """
+        schema = validators.Draft7Validator.META_SCHEMA
+        one = validators._RefResolver("", schema)
+        validator = validators.Draft7Validator(schema, resolver=one)
+        self.assertFalse(validator.is_valid({"maxLength": "foo"}))
+
+        another = {
+            "allOf": [{"$ref": validators.Draft7Validator.META_SCHEMA["$id"]}],
+        }
+        two = validators._RefResolver("", another)
+        validator = validators.Draft7Validator(another, resolver=two)
+        self.assertFalse(validator.is_valid({"maxLength": "foo"}))
+
+    def test_newly_created_validator_with_ref_resolver(self):
+        """
+        See https://github.com/python-jsonschema/jsonschema/issues/1061#issuecomment-1624266555.
+        """  # noqa: E501
+
+        def handle(uri):
+            self.assertEqual(uri, "http://example.com/foo")
+            return {"type": "integer"}
+
+        resolver = validators._RefResolver("", {}, handlers={"http": handle})
+        Validator = validators.create(
+            meta_schema={},
+            validators=validators.Draft4Validator.VALIDATORS,
+        )
+        schema = {"$id": "http://example.com/bar", "$ref": "foo"}
+        validator = Validator(schema, resolver=resolver)
+        self.assertEqual(
+            (validator.is_valid({}), validator.is_valid(37)),
+            (False, True),
+        )
+
+    def test_refresolver_with_pointer_in_schema_with_no_id(self):
+        """
+        See https://github.com/python-jsonschema/jsonschema/issues/1124#issuecomment-1632574249.
+        """  # noqa: E501
+
+        schema = {
+            "properties": {"x": {"$ref": "#/definitions/x"}},
+            "definitions": {"x": {"type": "integer"}},
+        }
+
+        validator = validators.Draft202012Validator(
+            schema,
+            resolver=validators._RefResolver("", schema),
+        )
+        self.assertEqual(
+            (validator.is_valid({"x": "y"}), validator.is_valid({"x": 37})),
+            (False, True),
+        )
+
+
 
 def sorted_errors(errors):
     def key(error):
@@ -2295,10 +2441,10 @@ def sorted_errors(errors):
     return sorted(errors, key=key)
 
 
-@attr.s
+@define
 class ReallyFakeRequests:
 
-    _responses = attr.ib()
+    _responses: dict[str, Any]
 
     def get(self, url):
         response = self._responses.get(url)
@@ -2307,10 +2453,10 @@ class ReallyFakeRequests:
         return _ReallyFakeJSONResponse(json.dumps(response))
 
 
-@attr.s
+@define
 class _ReallyFakeJSONResponse:
 
-    _response = attr.ib()
+    _response: str
 
     def json(self):
         return json.loads(self._response)
